/*******************************************************************************
 * ------------------ PROJECT : SunTris ----------------------------------------
 * File : SunTris.java
 * Author : Bastian Molkenthin  (nickname: Sunshine)
 * Project started on 12. Oktober 2004, 14:35
 * Last change : 19. November 2004
 * NOTE : coded for my own Nokia 3200 which display size is 128x96
 *        For other mobiles, hardcoded values have to be fixed!
 * TO DO: - good score system
 *        - update speed (=levels)  -> IN PROGRESS
 *        - pause option        -> DONE
 *        - help/button screen  -> DONE
 *        - ? options: different colors/parts ? -> WON'T BE DONE
 *        - splash screen       -> DONE
 *        - preview for next part   -> DONE
 *        - remove hardcoded values with calculated ones for portability -> WON'T BE DONE
 *        - collision detection from side  -> DONE
 *        - part movement must be improved -> IN PROGRESS
 ******************************************************************************/

package SunTris;

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import javax.microedition.rms.*;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import java.io.*;



public class SunTris extends MIDlet implements CommandListener {
    
    
    private Display display;    // The display for this MIDlet
    
    private List MainMenuList;
    private Form HelpForm;
    private Command HelpFormBackCommand;
    private Command chooseCommand;
    private Command exitMenuCommand;
    // GameScreen variables
    private GameScreen myGameScreen;
    private Command backGameCommand;
    private Command pauseGameCommand;
    private Command resumeGameCommand;
    // Splash screen
    private SplashScreen mySplashScreen;
    private static Image image;
    // HighScore
    private String[] highScoreNames = new String[3];
    private int[] highScorePoints = new int[3];
    private RecordStore highScoreDB; 
    private int numRecords;
    private Form HighScoreForm;
    
    
    public void startApp() {
        display = Display.getDisplay(this);
        
        mySplashScreen = createSplashScreen();
        display.setCurrent(mySplashScreen);
        
        createMainMenu();
        createHelpScreen();
        // load highScore database into array
        for (int i = 0; i < 3; i++) {
            highScoreNames[i] = "ddaf";
            highScorePoints[i] = 3;
        }
        //checkInitDatabase();
        //loadScore();
        //createHighScoreScreen();
        
        // create game screen
        myGameScreen = createGameScreen();
        backGameCommand = new Command("back", Command.BACK,0);
        pauseGameCommand = new Command("pause", Command.SCREEN,1);
        resumeGameCommand = new Command("resume", Command.SCREEN,1);
        myGameScreen.addCommand(backGameCommand);
        myGameScreen.addCommand(pauseGameCommand);
        
        myGameScreen.setCommandListener(this);
        
        //display.setCurrent(MainMenuList);
        
   
    }
    
   
    public void pauseApp() {
    }
    
   
    public void destroyApp(boolean unconditional) {
    }
    
    
    public void commandAction(Command c, Displayable d) {
        
        // Main Menu is visible
        if (d == MainMenuList){
            if (c == exitMenuCommand) {
                destroyApp(true);
                notifyDestroyed();
            } else {
            int index = MainMenuList.getSelectedIndex();
            
            if (index == 2) { // Exit Command
                destroyApp(true);
                notifyDestroyed();
            }
            
            if (index == 0) {
                display.setCurrent(myGameScreen);
            }
            
            if (index == 1) {  // Help screen
                display.setCurrent(HelpForm);
            }
            
            //if (index == 1) display.setCurrent(HighScoreForm);
            }
        }
        else
        // Game screen is visible
        if (d == myGameScreen) {
            if (c == backGameCommand) {
                display.setCurrent(MainMenuList);
            } else
            if (c == pauseGameCommand) {
                if (myGameScreen.running) myGameScreen.running = false; 
                    else  myGameScreen.running = true;
                myGameScreen.removeCommand(pauseGameCommand);        
                myGameScreen.addCommand(resumeGameCommand);
            } else
            if (c == resumeGameCommand) {
                myGameScreen.running = true;
                myGameScreen.removeCommand(resumeGameCommand);        
                myGameScreen.addCommand(pauseGameCommand);
                myGameScreen.resume();
                
            }
        }
        // Help screen is visible
        if (d == HelpForm) {
            if (c == HelpFormBackCommand) {
                display.setCurrent(MainMenuList);
            }
        }
        
        if (d == HighScoreForm) {
            if (c == HelpFormBackCommand) {
                //for (int i = 0; i < 3; i++)
                //    highScoreNames[i] = "MR X";
                //saveDatabase();
                display.setCurrent(MainMenuList);
            }
        }
    }
    
    
    // ** Create Main Menu
    public void createMainMenu() {
        MainMenuList = new List("Main Menu",List.IMPLICIT);
        MainMenuList.append("Play Game",null);
        //MainMenuList.append("HighScore",null);
        MainMenuList.append("Help",null);
        MainMenuList.append("Exit", null);
        chooseCommand = new Command("Choose", Command.OK, 0);
        exitMenuCommand = new Command("Exit", Command.BACK, 1);
        MainMenuList.addCommand(chooseCommand);
        MainMenuList.addCommand(exitMenuCommand);
        MainMenuList.setCommandListener(this);            
    }
    
    // ** Create help screen
    public void createHelpScreen() {
        HelpForm = new Form(" - Suntris Help - ");
        HelpForm.append("Left : Key 4\n");
        HelpForm.append("Right : Key 6\n");
        HelpForm.append("Rotate #1 : Key 2\n");
        HelpForm.append("Rotate #2 : Key 8\n");
        HelpForm.append("Move down : Key 5\n");
        HelpForm.append("Drop part : Key 3\n");
        HelpFormBackCommand = new Command("back", Command.BACK, 1);
        HelpForm.addCommand(HelpFormBackCommand);
        HelpForm.setCommandListener(this);
    }
    
    // ** Create HighScore screen
    public void createHighScoreScreen () {
        HighScoreForm = new Form(" - HighScores -");
        //HighScoreForm.append(Integer.toString(numRecords));
        for (int i = 0; i < 3; i++)
            HighScoreForm.append(highScoreNames[i]);
        HighScoreForm.addCommand(HelpFormBackCommand);
        HighScoreForm.setCommandListener(this);
    }
    
    
    // ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
    // ***************** DATABASE STUFF ***************************************
    // ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
    /*final int MAX_REC_LEN = 16;
    StringBuffer recBuf = new StringBuffer(MAX_REC_LEN);
    byte[] data;
    
    protected void checkInitDatabase() {
        try {
            highScoreDB = RecordStore.openRecordStore("HS", false);
            highScoreDB.closeRecordStore();
        }
        catch (RecordStoreException ex) {
            initDatabase();
        }
    }
    
    protected void initDatabase() {
        try {
            highScoreDB = RecordStore.openRecordStore("HS", true);
            clearBuf();
            for (int i = 0; i < 3; i++) {
                data = stringToBufferToBytes("default", 0);
                highScoreDB.addRecord(data, 0, data.length);
                clearBuf();
            }
            highScoreDB.closeRecordStore();
        }
        catch(RecordStoreException ex) { }
    }
    
    protected void resetDatabase() {
        try {
            highScoreDB.deleteRecordStore("HS");
            initDatabase();
        }
        catch (RecordStoreNotFoundException ex) {
            initDatabase();
        }
        catch(RecordStoreException e) { }
    }
    
    protected void loadScore() {
        byte[] myScore = new byte[50];
    
        try {
            highScoreDB = RecordStore.openRecordStore("HS", false);
            for (int i = 1; i < 4; i++)
            {
                highScoreDB.getRecord(i, myScore, 0);
                highScoreNames[i-1] = myScore.toString();
            }
        }
        catch(RecordStoreException e) { }
    }
    
    protected void saveScore(String name, int score) {
        byte[] data = stringToBufferToBytes(name, score);
        byte[] oldScore = new byte[50];
        try {
            highScoreDB = RecordStore.openRecordStore("HS", false);
            for (int i = 4; i > 0; i--) {
                highScoreDB.getRecord(i, oldScore, 0);
                if (score > Integer.parseInt(new String(oldScore, 7, 8).trim())) {
                    highScoreDB.setRecord(i + 1, oldScore, 0, oldScore.length);
                    if (i == 1) {
                        highScoreDB.setRecord(i, data, 0, data.length);
                    }
                } else
                {
                    highScoreDB.setRecord(i + 1, data, 0, data.length);
                    break;
                }
            }
            data = null;
            highScoreDB.closeRecordStore();
        }
        catch (RecordStoreException e) { }
    }
    
    protected void clearBuf() {
        for (int i = 0; i < MAX_REC_LEN; i++) {
            recBuf.insert(i,' ');
        }
        recBuf.setLength(MAX_REC_LEN);
    }
    
    protected byte[] stringToBufferToBytes(String name, int score) {
        clearBuf();
        recBuf.insert(0, name);
        //recBuf.insert(7, Integer.toString(score));
        return recBuf.toString().getBytes();
    }
    
    */
    
    
    
    // ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
    // ***************** SPLASH SCREEN STUFF **********************************
    // ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
    
    public SplashScreen createSplashScreen() {
        return new SplashScreen();
    }
       
    class SplashScreen extends Canvas {
        protected Timer splashtimer;
        protected TimerTask timertask;
        
        protected void showNotify(){
            startSplashTimer();
            repaint();
        }
        protected void hideNotify(){
            stopSplashTimer();
        }
        
        protected void stopSplashTimer() {
            splashtimer.cancel();
        }
        
        protected void startSplashTimer() {
            splashtimer = new Timer();
            timertask = new TimerTask() {
                public void run() {
                    display.setCurrent(MainMenuList);          
                }
            };
            splashtimer.schedule(timertask,1500);
        }
        
        protected void paint(Graphics g) {
            int width = getWidth();
            int height = getHeight();
            g.setColor(0x00FFFFFF);
            g.fillRect(0, 0, width, height);
            if (image == null) {
                try {
                    image = Image.createImage("/SunTris/resources/splash.png");
                } catch (IOException ex) {
                    image = null;
                }
            }
            g.drawImage(image, width/2, height/2, Graphics.VCENTER | Graphics.HCENTER);
        }
        
    }
    
    // -------   ** SPLASH SCREEN STUFF  :: END :: **    ------
    
    
    
    
    // :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
    // -----------------   ** MAIN GAME STUFF **   ------------------ 
    // ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: 
    
    public GameScreen createGameScreen() {
        return new GameScreen();
    }
    
    class GameScreen extends Canvas {
        
        protected int width = getWidth();
        protected int height = getHeight();
        private boolean running = false;
        private final int DELAY = 100;
        private GameThread myGameThread;
        private int key;                // key last pressed
        // tray is 12 blocks wide and 18 blocks high
        private Block[][] blocks = new Block[12][18];
        // increasing var -> equals MoveDownDelay -> part moves one down
        private int counterToMoveDown;
        private int MoveDownDelay = 10;
        private TetPart curPart;
        private TetPart nextPart;
        //private boolean bGameOver;
        private int GameOverTimeOut;  // > 0 -> gameover
        private Random mainRandom;
        private int score;
        
        // :: constructor
        GameScreen() {
            mainRandom = new Random();
            
        }
        
        
        // :: when GameScreen becomes visible
        protected void showNotify() {
            counterToMoveDown = 0;
            // initialize all blocks empty
            for (int i = 0; i < 12; i++)
                for (int j = 0; j < 18; j++)
                    blocks[i][j] = new Block();
            
            running = true;
            GameOverTimeOut = 0;
            score = 0;
            curPart = createNewPart();
            nextPart = createNewPart();
            myGameThread = new GameThread();
            myGameThread.start();
            
    
        }
        
        protected void resume() {
            myGameThread = null;
            myGameThread = new GameThread();
            myGameThread.start();
        }
        
        // :: when GameScreen becomes invisible
        protected void hidenotify() {
            running = false;
        }
        
        
        protected void keyPressed(int keycode) {
            key = keycode;
           
        }
        
        // :: creates new random part
        protected TetPart createNewPart() {
            int temp = (mainRandom.nextInt() >>> 1) % 6;
            switch (temp) {
                case 0 : return new T_Part(); 
                case 1 : return new L_Part(); 
                case 2 : return new O_Part(); 
                case 3 : return new Z_Part(); 
                case 4 : return new Z2_Part(); 
                case 5 : return new I_Part(); 
                default : return new T_Part();
            } // end switch
            
        } // :: end createNewPart()
        
        
        // :: main game logic
        protected void DoGameStuff() {
            
            if (++counterToMoveDown > MoveDownDelay) {      // time to move down one block
                counterToMoveDown = 0;
                
                curPart.moveOneDown();
                if (key != 0) curPart.moveRot(key); // left, right, rotate
                curPart.checkCol();
                key = 0;
                
                //** if one row completed delete it
                // check every row
                for (int j = 0; j < 18; j++) { 
                    if ( blocks[0][j].full && blocks[1][j].full && blocks[2][j].full &&
                         blocks[3][j].full && blocks[4][j].full && blocks[5][j].full &&
                         blocks[6][j].full && blocks[7][j].full && blocks[8][j].full &&
                         blocks[9][j].full && blocks[10][j].full && blocks[11][j].full ) {
                             // ok one row full, so move all blocks above one block down
                             for (int h = 0; h < 12; h++) {
                                for (int i = j; i > 0; i--) {
                                    blocks[h][i].color = blocks[h][i-1].color;
                                    blocks[h][i].full = blocks[h][i-1].full;
                                }
                                // if one row removed, top row is empty for sure
                                blocks[h][0].full = false;
                             }
                             // increase score
                             score += 100 + 10 * (18-j);
                             // check if we reached next level... DOES IT WORK???
                             if (score % 500 == 0) MoveDownDelay = 10 - (score % 500);
                            
                    } // end if             
                } // end for-loop row
            //
            } // end if counterToMoveDown
            else {
                if (key != 0) { curPart.moveRot(key); key = 0;}
            }
            
            
        } // end DoGameStuff
        
        
        
        // ** paint method
        protected void paint (Graphics g) {
            
            // clear background with white
            g.setColor(255,255,255);
            g.fillRect(0, 0, width, height);
            
            // if GameOver, print "GameOver"
            if (GameOverTimeOut > 0) {
                GameOverTimeOut++;
                Font font = g.getFont();
                font = Font.getFont(Font.FACE_PROPORTIONAL, 0, Font.SIZE_SMALL);
                g.setFont(font);
                g.setColor(0x00000000);
                g.drawString("Game over!", width / 2 , height / 2 - 10, 
                                Graphics.HCENTER | Graphics.TOP);
                String str = "Score : " + Integer.toString(score);
                g.drawString(str, width / 2 , height / 2 + 10, 
                                Graphics.HCENTER | Graphics.TOP);
                // game over screen is over, move to mainmenu
                if (GameOverTimeOut > 20) {    
                    GameOverTimeOut = 0;    
                    commandAction(backGameCommand, myGameScreen);
                }
                return;
            }
            
            // paint "tray" |_|
            g.setColor(0,0,0);
            g.drawLine(10 - 1,5,10 - 1 ,95);
            g.drawLine(70,5,70,95);
            g.drawLine(10,95,70,95);
            
            // draw score
            Font font = g.getFont();
            font = Font.getFont(Font.FACE_PROPORTIONAL, 0, Font.SIZE_SMALL);
            g.setFont(font);
            g.drawString("Score:", width - 30, height / 3 - height / 4, 
                                Graphics.HCENTER | Graphics.TOP);
            g.drawString(Integer.toString(score), width - 30, height / 3 - height / 4 + 10, 
                                Graphics.HCENTER | Graphics.TOP);
            
            // if game is paused, draw it
            if (running == false) {
                g.setColor(0x00FF0000);
                g.drawString("Game paused",width / 2 - 20, height / 2, 
                                Graphics.HCENTER | Graphics.TOP);
                
            }
                                
            
            g.setColor(0x00000000);
            // draw next part
            g.drawString("Next:", width - 30, height / 2 - 10, 
                                Graphics.HCENTER | Graphics.TOP);   
            g.setColor(nextPart.color);
            if (nextPart instanceof T_Part) {
                g.fillRect(width - 30, height / 2 + 5, 5*3,5);
                g.fillRect(width - 30 + 5, height / 2 + 10, 5,5);
            }
            if (nextPart instanceof L_Part) {
                g.fillRect(width - 30, height / 2 + 5, 5*3,5);
                g.fillRect(width - 30 + 10, height / 2 + 10, 5,5);
            }
            if (nextPart instanceof O_Part) {
                g.fillRect(width - 30, height / 2 + 5, 5*2,5*2);
            }
            if (nextPart instanceof Z_Part) {
                g.fillRect(width - 30, height / 2 + 5, 5*2,5);
                g.fillRect(width - 30 + 5, height / 2 + 10, 5*2,5);
            }
            if (nextPart instanceof Z2_Part) {
                g.fillRect(width - 30 + 5, height / 2 + 5, 5*2,5);
                g.fillRect(width - 30, height / 2 + 10, 5*2,5);
            }
            if (nextPart instanceof I_Part) {
                g.fillRect(width - 30, height / 2 + 5, 5*4,5);
            }
            
            // paint all blocks
            for (int i = 0; i < 12; i++)
                for (int j = 0; j < 18; j++) {
                    if (blocks[i][j].full) {
                        g.setColor(blocks[i][j].color);
                        g.fillRect(10 + i*5, 5 + j*5, 5, 5);
                    }
                }
            
            // paint current part
            if (curPart == null) return;
            g.setColor(curPart.color);
            for (int i = 0; i < 4; i++){
                g.fillRect(10 + curPart.blocksX[i] * 5, 5 + curPart.blocksY[i] * 5, 5, 5);
            }
            
            
            
        } // :: end paint()
        
        public class GameThread extends Thread {
            public void run() {
                while (running) {
                    long time = System.currentTimeMillis();
                                              
                    
                    DoGameStuff();
                    
                    //key = 0;
                    repaint();
                    
                    time = System.currentTimeMillis() - time;
                    try
                    {
                        if (time < DELAY) {
                            Thread.sleep(DELAY - (int)time);
                        }
                    }
                    catch (Exception ex) { }
                } // end while
                if (running == false) repaint();
            } // end run
            
        } // end GameThread
        
        
        
        
        
        
        
        // ************** Class representing a part ****************************
        // ** abstract; objects to initiate heritate from TetPart
        // *********************************************************************
        abstract class TetPart {
            
            int color;
            int rotated;
            int[] blocksX; int[] blocksY;
            Random random;
            
            // ** constructor
            TetPart() {
                blocksX = new int[4];
                blocksY = new int[4];
                rotated = 0;
            
                // color
                random = new Random();
                int tempint = (random.nextInt() >>> 1) % 7; 
                switch(tempint) {
                    case 0 : color = 0x0000FF00; break;  // green
                    case 1 : color = 0x000000FF; break;  // blue
                    case 2 : color = 0x00FFFF00; break;  // yellow
                    case 3 : color = 0x00FF0000; break;  // red
                    case 4 : color = 0x00FF00FF; break;
                    case 5 : color = 0x0000AA00; break;
                    case 6 : color = 0x0000AAFF; break;
                    default : color = 0x000000FF;
                }
            } // :: end constructor
        
            // ** moveOneDown()  : moves currentPart one block down
            // returns false is part reached floor or top of another part,
            // in this case a new part is automatically created;
            // returns true otherwise
            void moveOneDown() {
                // check if Game Over 
                if ( (curPart.blocksY[0] == 0 && blocks[curPart.blocksX[0]][1].full) ||
                     (curPart.blocksY[1] == 0 && blocks[curPart.blocksX[1]][1].full) ||
                     (curPart.blocksY[2] == 0 && blocks[curPart.blocksX[2]][1].full) ||
                     (curPart.blocksY[3] == 0 && blocks[curPart.blocksX[3]][1].full) ) {
                        GameOverTimeOut = 1;    
                }
                
                // move part one block down
                for (int i = 0; i < 4; i++) curPart.blocksY[i] += 1;
            } // :: end moveOneDown
            
            boolean checkCol() {
                
                
                // check if part is now on floor or on top of another part
                if (curPart.blocksY[0] == 17 || curPart.blocksY[1] == 17 ||
                    curPart.blocksY[2] == 17 || curPart.blocksY[3] == 17 ||
                    blocks[curPart.blocksX[0]][curPart.blocksY[0]+1].full == true ||
                    blocks[curPart.blocksX[1]][curPart.blocksY[1]+1].full == true ||
                    blocks[curPart.blocksX[2]][curPart.blocksY[2]+1].full == true ||
                    blocks[curPart.blocksX[3]][curPart.blocksY[3]+1].full == true ) {
                        
                                blocks[curPart.blocksX[0]][curPart.blocksY[0]].full = true;
                                blocks[curPart.blocksX[1]][curPart.blocksY[1]].full = true;
                                blocks[curPart.blocksX[2]][curPart.blocksY[2]].full = true;
                                blocks[curPart.blocksX[3]][curPart.blocksY[3]].full = true;
                                blocks[curPart.blocksX[0]][curPart.blocksY[0]].color = curPart.color;
                                blocks[curPart.blocksX[1]][curPart.blocksY[1]].color = curPart.color;
                                blocks[curPart.blocksX[2]][curPart.blocksY[2]].color = curPart.color;
                                blocks[curPart.blocksX[3]][curPart.blocksY[3]].color = curPart.color;
                                curPart = nextPart;
                                nextPart = null;
                                nextPart = createNewPart(); 
                                return false;
                } // end if
                else 
                    return true;
            
            } // ::  end checkCol
        
            // ** moveRot()  : handles key-presses
            // needs to be implemented in heritated classes
            abstract void moveRot(int key);
        
    } // :: end class TetPart
        
    
        
    // *************** class representing one block of a part *****************
    class Block {
        int color;
        boolean full;
        
        Block() {
            full = false;
        }
    } // :: end class Block
      
    
    
    
    
    
    // *************************************************************************  
    // ****************** Custom parts deriving TetPart ************************
    // *************************************************************************
    
    private final int Key_Left = KEY_NUM4;
    private final int Key_Right = KEY_NUM6;
    private final int Key_RotateCW = KEY_NUM2;
    private final int Key_RotateCCW = KEY_NUM8;
    private final int Key_MoveOneDown = KEY_NUM5;
    private final int Key_MoveDown = KEY_NUM3;
    
    //   _____
    //  |_   _|  - Part class ********************************************
    //    |_|    
        
    class T_Part extends TetPart {
        
        // ** constructor
        T_Part() {
            super();
            blocksX[0] = 5; blocksY[0] = 0;
            blocksX[1] = 6; blocksY[1] = 0;
            blocksX[2] = 7; blocksY[2] = 0;
            blocksX[3] = 6; blocksY[3] = 1;
        }
        
        // ** implementation of moverot() 
        void moveRot(int key) {
            switch (key) {
                
                // drop part
                case Key_MoveDown: { 
                    while ( checkCol() ) { moveOneDown(); };     
                } break;
                
                case Key_Left: {  // move left
                    if (blocksX[0] != 0 && blocksX[1] != 0 &&  // not on left side
                        blocksX[2] != 0 && blocksX[3] != 0 &&
                        blocks[blocksX[0]-1][blocksY[0]].full == false && // no other part left of part
                        blocks[blocksX[1]-1][blocksY[1]].full == false &&
                        blocks[blocksX[2]-1][blocksY[2]].full == false &&
                        blocks[blocksX[3]-1][blocksY[3]].full == false) {
                        for (int i = 0; i < 4; i++) blocksX[i] -= 1;
                    }     
                } break;
               
                case Key_Right: {  // move right
                    if (blocksX[0] != 11 && blocksX[1] != 11 && blocksX[2] != 11 && blocksX[3] != 11 &&
                        blocks[blocksX[0]+1][blocksY[0]].full == false && // no other part left of part
                        blocks[blocksX[1]+1][blocksY[1]].full == false &&
                        blocks[blocksX[2]+1][blocksY[2]].full == false &&
                        blocks[blocksX[3]+1][blocksY[3]].full == false) {
                        for (int i = 0; i < 4; i++) blocksX[i] += 1;
                    }     
                } break;
                
                case Key_MoveOneDown: {
                    moveOneDown();
                } break;
               
                
                case Key_RotateCW: {
                    rotated += 90; if (rotated == 360) rotated = 0;
                    
                        switch(rotated) {
                            case 90 : { 
                                blocksX[2] = blocksX[3]; blocksY[2] = blocksY[3];
                                blocksX[3] = blocksX[0]; blocksY[3] = blocksY[0];
                                blocksX[0] = blocksX[1]; blocksY[0] = blocksY[1] -1;
                            } break;
                            
                            case 180 : {
                                blocksX[0] = blocksX[3]; blocksY[0] = blocksY[3];
                                blocksX[3] = blocksX[1]; blocksY[3] = blocksY[1] - 1;
                                blocksX[2] = blocksX[1] + 1; blocksY[2] = blocksY[1];
                            } break;
                            
                            case 270 : {
                                blocksX[0] = blocksX[3]; blocksY[0] = blocksY[3];
                                blocksX[3] = blocksX[2]; blocksY[3] = blocksY[2];
                                blocksX[2] = blocksX[1]; blocksY[2] = blocksY[1] + 1;  
                            } break;
                            
                            case 0 : {
                                blocksX[2] = blocksX[3]; blocksY[2] = blocksY[3];
                                blocksX[3] = blocksX[1]; blocksY[3] = blocksY[1] + 1;
                                blocksX[0] = blocksX[1] - 1; blocksY[0] = blocksY[1];
                            }
                        } // end switch rotated
                    
                    
                    // check if blocks are outside due to rotation and correct it
                    for (int i = 0; i < 4; i++) {
                        if (blocksX[i] < 0) 
                            for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] + 1;
                        else
                            if (blocksX[i] > 11)
                                for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] - 1;    
                    }
                } break;
                
                
                case Key_RotateCCW: {
                    rotated -= 90; if (rotated == -90) rotated = 270;
                    
                        switch(rotated) {
                            case 90 : { 
                                blocksX[0] += 1; blocksY[0] -= 1;
                                blocksX[2] -= 1; blocksY[2] += 1;
                                blocksX[3] -= 1; blocksY[3] += 1;
                            } break;
                            
                            case 180 : {
                                blocksX[0] -= 1; blocksY[0] += 1;
                                blocksX[2] += 1; blocksY[2] -= 1;
                                blocksX[3] -= 1; blocksY[3] -= 1;
                            } break;
                            
                            case 270 : {
                                blocksX[0] += 1; blocksY[0] -= 1;
                                blocksX[2] -= 1; blocksY[2] += 1;
                                blocksX[3] += 1; blocksY[3] -= 1; 
                            } break;
                            
                            case 0 : {
                                blocksX[0] -= 1; 
                                blocksY[1] -= 1;
                                blocksX[2] += 1; blocksY[2] -= 2; 
                                blocksX[3] += 1;
                            }
                        } // end switch rotated
                    
                    
                    // check if blocks are outside due to rotation and correct it
                    for (int i = 0; i < 4; i++) {
                        if (blocksX[i] < 0) 
                            for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] + 1;
                        else
                            if (blocksX[i] > 11)
                                for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] - 1;    
                    }
                } break;
                
            } // end switch(key)
            
        } // :: end moveRot      
       
        
    } // :: end T_Part class
    
    
    
    //  ____
    // |    |  - Part class **********************************************
    // |____|   
         
    class O_Part extends TetPart {
        
        // ** constructor
        O_Part() {
            super();
            blocksX[0] = 5; blocksY[0] = 0;
            blocksX[1] = 6; blocksY[1] = 0;
            blocksX[2] = 5; blocksY[2] = 1;
            blocksX[3] = 6; blocksY[3] = 1;
        } 
        
        // ** custom implentation of moveRot()
        void moveRot(int key) {
            switch (key) {
                
                // drop part
                case Key_MoveDown: { 
                    while ( checkCol() ) { moveOneDown(); };    
                } break;
                
                case Key_Left: {  // move left
                    if (blocksX[0] != 0 && blocksX[1] != 0 && blocksX[2] != 0 && blocksX[3] != 0 &&
                        blocks[blocksX[0]-1][blocksY[0]].full == false && // no other part left of part
                        blocks[blocksX[1]-1][blocksY[1]].full == false &&
                        blocks[blocksX[2]-1][blocksY[2]].full == false &&
                        blocks[blocksX[3]-1][blocksY[3]].full == false) {
                        for (int i = 0; i < 4; i++) blocksX[i] -= 1;
                    }     
                } break;
               
                case Key_Right: {  // move right
                    if (blocksX[0] != 11 && blocksX[1] != 11 && blocksX[2] != 11 && blocksX[3] != 11 &&
                        blocks[blocksX[0]+1][blocksY[0]].full == false && // no other part left of part
                        blocks[blocksX[1]+1][blocksY[1]].full == false &&
                        blocks[blocksX[2]+1][blocksY[2]].full == false &&
                        blocks[blocksX[3]+1][blocksY[3]].full == false) {
                        for (int i = 0; i < 4; i++) blocksX[i] += 1;
                    }     
                } break;
                
                case Key_MoveOneDown: {
                    moveOneDown();
                } break;
               
                                
            } // end switch(key)
        }
    } // :: end O_Part class
    
    
    
    //  _
    // | |
    // | |_  - Part class ****************************************************
    // |___|
    class L_Part extends TetPart {
        
        // ** constructor
        L_Part() {
            super();
            blocksX[0] = 5; blocksY[0] = 0;
            blocksX[1] = 6; blocksY[1] = 0;
            blocksX[2] = 7; blocksY[2] = 0;
            blocksX[3] = 7; blocksY[3] = 1;
        }
        
        // ** custom implentation of moveRot()
        void moveRot(int key) {
            switch (key) {
                
                // drop part
                case Key_MoveDown: { 
                    while ( checkCol() ) { moveOneDown(); };              
                } break;
                
                case Key_Left: {  // move left
                    if (blocksX[0] != 0 && blocksX[1] != 0 && blocksX[2] != 0 && blocksX[3] != 0 &&
                        blocks[blocksX[0]-1][blocksY[0]].full == false && // no other part left of part
                        blocks[blocksX[1]-1][blocksY[1]].full == false &&
                        blocks[blocksX[2]-1][blocksY[2]].full == false &&
                        blocks[blocksX[3]-1][blocksY[3]].full == false) {
                        for (int i = 0; i < 4; i++) blocksX[i] -= 1;
                    }     
                } break;
               
                case Key_Right: {  // move right
                    if (blocksX[0] != 11 && blocksX[1] != 11 && blocksX[2] != 11 && blocksX[3] != 11 &&
                        blocks[blocksX[0]+1][blocksY[0]].full == false && // no other part left of part
                        blocks[blocksX[1]+1][blocksY[1]].full == false &&
                        blocks[blocksX[2]+1][blocksY[2]].full == false &&
                        blocks[blocksX[3]+1][blocksY[3]].full == false) {
                        for (int i = 0; i < 4; i++) blocksX[i] += 1;
                    }     
                } break;
                
                case Key_MoveOneDown: {
                    moveOneDown();
                } break;
                
                case Key_RotateCW: {
                    rotated += 90; if (rotated == 360) rotated = 0;
                    
                        switch(rotated) {
                            case 90 : { 
                                blocksX[0] +=1;
                                blocksY[1] +=1;
                                blocksX[2] -=1; blocksY[2] +=2;
                                blocksX[3] -=2; blocksY[3] +=1;
                            } break;
                            
                            case 180 : {
                                blocksX[0] += 1; blocksY[0] += 1;
                                blocksX[2] -= 1; blocksY[2] -= 1;
                                blocksY[3] -= 2;
                            } break;
                            
                            case 270 : {
                                blocksX[0] -= 1; blocksY[0] += 1;
                                blocksX[2] += 1; blocksY[2] -= 1;
                                blocksX[3] += 2;
                            } break;
                            
                            case 0 : {
                                blocksX[0] -= 1; blocksY[0] -= 2;
                                blocksY[1] -= 1;
                                blocksX[2] += 1;
                                blocksY[3] += 1;
                            }
                        } // end switch rotated
                    
                    
                    // check if blocks are outside due to rotation and correct it
                    for (int i = 0; i < 4; i++) {
                        if (blocksX[i] < 0) 
                            for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] + 1;
                        else
                            if (blocksX[i] > 11)
                                for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] - 1;    
                    }
                } break;
               
                // Rotate counter-clockwise
                case Key_RotateCCW: {
                    rotated -= 90; if (rotated == -90) rotated = 270;
                    
                        switch(rotated) {
                            case 90 : { 
                                blocksX[0] -= 1; blocksY[0] -= 1;
                                blocksX[2] += 1; blocksY[2] += 1;
                                blocksY[3] += 2;
                            } break;
                            
                            case 180 : {
                                blocksX[0] += 1; blocksY[0] -= 1;
                                blocksX[2] -= 1; blocksY[2] += 1;
                                blocksX[3] -= 2;
                            } break;
                            
                            case 270 : {
                                blocksX[0] += 1; blocksY[0] += 2;
                                blocksY[1] += 1;
                                blocksX[2] -= 1;
                                blocksY[3] -= 1;
                            } break;
                            
                            case 0 : {
                                blocksX[0] -= 1;
                                blocksY[1] -= 1;
                                blocksX[2] += 1; blocksY[2] -= 2;
                                blocksX[3] += 2; blocksY[3] -= 1;
                            }
                        } // end switch rotated
                    
                    
                    // check if blocks are outside due to rotation and correct it
                    for (int i = 0; i < 4; i++) {
                        if (blocksX[i] < 0) 
                            for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] + 1;
                        else
                            if (blocksX[i] > 11)
                                for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] - 1;    
                    }
                } break;
                                
            } // end switch(key)
        } // :: end moveRot()
        
    } // :: end class L_Part
    
    
    
    //   _____
    //  |__  |__ - part class ************************************************
    //     |____|
    
    class Z_Part extends TetPart {
        
        // constructor
        Z_Part() {
            super();
            blocksX[0] = 5; blocksY[0] = 0;
            blocksX[1] = 6; blocksY[1] = 0;
            blocksX[2] = 6; blocksY[2] = 1;
            blocksX[3] = 7; blocksY[3] = 1;
        }
        
        // ** custom implentation of moveRot()
        void moveRot(int key) {
            switch (key) {
                
                // drop part
                case Key_MoveDown: { 
                    while ( checkCol() ) { moveOneDown(); };               
                } break;
                
                case Key_Left: {  // move left
                    if (blocksX[0] != 0 && blocksX[1] != 0 && blocksX[2] != 0 && blocksX[3] != 0 &&
                        blocks[blocksX[0]-1][blocksY[0]].full == false && // no other part left of part
                        blocks[blocksX[1]-1][blocksY[1]].full == false &&
                        blocks[blocksX[2]-1][blocksY[2]].full == false &&
                        blocks[blocksX[3]-1][blocksY[3]].full == false) {
                        for (int i = 0; i < 4; i++) blocksX[i] -= 1;
                    }     
                } break;
               
                case Key_Right: {  // move right
                    if (blocksX[0] != 11 && blocksX[1] != 11 && blocksX[2] != 11 && blocksX[3] != 11 &&
                        blocks[blocksX[0]+1][blocksY[0]].full == false && // no other part right of part
                        blocks[blocksX[1]+1][blocksY[1]].full == false &&
                        blocks[blocksX[2]+1][blocksY[2]].full == false &&
                        blocks[blocksX[3]+1][blocksY[3]].full == false) {
                        for (int i = 0; i < 4; i++) blocksX[i] += 1;
                    }     
                } break;
                
                case Key_MoveOneDown: {
                    moveOneDown();
                } break;
                
                case Key_RotateCCW:
                case Key_RotateCW: {
                    rotated += 90; if (rotated == 180) rotated = 0;
                    
                        switch(rotated) {
                            case 90 : { 
                                blocksX[0] +=1;
                                blocksY[1] +=1;
                                blocksX[2] -=1;
                                blocksX[3] -=2; blocksY[3] +=1;
                            } break;
                            
                            case 0 : {
                                blocksX[0] -= 1;
                                blocksY[1] -= 1;
                                blocksX[2] += 1; 
                                blocksX[3] += 2; blocksY[3] -=1;
                            }
                        } // end switch rotated
                    
                    
                    // check if blocks are outside due to rotation and correct it
                    for (int i = 0; i < 4; i++) {
                        if (blocksX[i] < 0) 
                            for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] + 1;
                        else
                            if (blocksX[i] > 11)
                                for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] - 1;    
                    }
                } break;
                
                                                
            } // end switch(key)
        } // :: end moveRot()
    } // :: end Z_Part class
    
    
    
    
    //     _____
    //  __|   __| - part class ************************************************
    // |_____|
    
    class Z2_Part extends TetPart {
        
        // constructor
        Z2_Part() {
            super();
            blocksX[0] = 5; blocksY[0] = 1;
            blocksX[1] = 6; blocksY[1] = 1;
            blocksX[2] = 6; blocksY[2] = 0;
            blocksX[3] = 7; blocksY[3] = 0;
        }
        
        // ** custom implentation of moveRot()
        void moveRot(int key) {
            switch (key) {
                
                // drop part
                case Key_MoveDown: { 
                    while ( checkCol() ) { moveOneDown(); };               
                } break;
                
                case Key_Left: {  // move left
                    if (blocksX[0] != 0 && blocksX[1] != 0 && blocksX[2] != 0 && blocksX[3] != 0 &&
                        blocks[blocksX[0]-1][blocksY[0]].full == false && // no other part left of part
                        blocks[blocksX[1]-1][blocksY[1]].full == false &&
                        blocks[blocksX[2]-1][blocksY[2]].full == false &&
                        blocks[blocksX[3]-1][blocksY[3]].full == false) {
                        for (int i = 0; i < 4; i++) blocksX[i] -= 1;
                    }     
                } break;
               
                case Key_Right: {  // move right
                    if (blocksX[0] != 11 && blocksX[1] != 11 && blocksX[2] != 11 && blocksX[3] != 11 &&
                        blocks[blocksX[0]+1][blocksY[0]].full == false && // no other part left of part
                        blocks[blocksX[1]+1][blocksY[1]].full == false &&
                        blocks[blocksX[2]+1][blocksY[2]].full == false &&
                        blocks[blocksX[3]+1][blocksY[3]].full == false) {
                        for (int i = 0; i < 4; i++) blocksX[i] += 1;
                    }     
                } break;
                
                case Key_MoveOneDown: {
                    moveOneDown();
                } break;
                
                case Key_RotateCCW:
                case Key_RotateCW: {
                    rotated += 90; if (rotated == 180) rotated = 0;
                    
                        switch(rotated) {
                            case 90 : { 
                                blocksY[0] -=1;
                                blocksX[1] -=1;
                                blocksY[2] +=1;
                                blocksX[3] -=1; blocksY[3] +=2;
                            } break;
                            
                            case 0 : {
                                blocksY[0] +=1;
                                blocksX[1] +=1;
                                blocksY[2] -=1;
                                blocksX[3] +=1; blocksY[3] -=2;
                            }
                        } // end switch rotated
                    
                    
                    // check if blocks are outside due to rotation and correct it
                    for (int i = 0; i < 4; i++) {
                        if (blocksX[i] < 0) 
                            for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] + 1;
                        else
                            if (blocksX[i] > 11)
                                for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] - 1;    
                    }
                } break;
               
                                
            } // end switch(key)
        } // :: end moveRot()
    } // :: end Z2_Part class
    
    
    
    //  __
    // |  |
    // |  | - part class *****************************************************
    // |  |
    // |__|
    
    class I_Part extends TetPart {
        
        // constructor
        I_Part() {
            super();
            blocksX[0] = 4; blocksY[0] = 0;
            blocksX[1] = 5; blocksY[1] = 0;
            blocksX[2] = 6; blocksY[2] = 0;
            blocksX[3] = 7; blocksY[3] = 0;
        }
        
        // ** custom implentation of moveRot()
        void moveRot(int key) {
            switch (key) {
                
                // drop part
                case Key_MoveDown: { 
                    while ( checkCol() ) { moveOneDown(); };              
                } break;
                
                case Key_Left: {  // move left
                    if (blocksX[0] != 0 && blocksX[1] != 0 && blocksX[2] != 0 && blocksX[3] != 0 &&
                        blocks[blocksX[0]-1][blocksY[0]].full == false && // no other part left of part
                        blocks[blocksX[1]-1][blocksY[1]].full == false &&
                        blocks[blocksX[2]-1][blocksY[2]].full == false &&
                        blocks[blocksX[3]-1][blocksY[3]].full == false) {
                        for (int i = 0; i < 4; i++) blocksX[i] -= 1;
                    }     
                } break;
               
                case Key_Right: {  // move right
                    if (blocksX[0] != 11 && blocksX[1] != 11 && blocksX[2] != 11 && blocksX[3] != 11 &&
                        blocks[blocksX[0]+1][blocksY[0]].full == false && // no other part left of part
                        blocks[blocksX[1]+1][blocksY[1]].full == false &&
                        blocks[blocksX[2]+1][blocksY[2]].full == false &&
                        blocks[blocksX[3]+1][blocksY[3]].full == false) {
                        for (int i = 0; i < 4; i++) blocksX[i] += 1;
                    }     
                } break;
                
                case Key_MoveOneDown: {
                    moveOneDown();
                } break;
                
                case Key_RotateCCW:
                case Key_RotateCW: {
                    rotated += 90; if (rotated == 180) rotated = 0;
                    
                        switch(rotated) {
                            case 90 : { 
                                blocksX[0] +=1;
                                blocksY[1] +=1;
                                blocksX[2] -=1; blocksY[2] +=2;
                                blocksX[3] -=2; blocksY[3] +=3;
                            } break;
                            
                            case 0 : {
                                blocksX[0] -=1;
                                blocksY[1] -=1;
                                blocksX[2] +=1; blocksY[2] -=2;
                                blocksX[3] +=2; blocksY[3] -=3;
                            }
                        } // end switch rotated
                    
                    
                    // check if blocks are outside due to rotation and correct it
                    for (int i = 0; i < 4; i++) {
                        if (blocksX[i] < 0) 
                            for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] + 1;
                        else
                            if (blocksX[i] > 11)
                                for (int j = 0; j < 4; j++) blocksX[j] = blocksX[j] - 1;    
                    }
                } break;
               
                                
            } // end switch(key)
        } // :: end moveRot()
    } // :: end I_Part class
    
    
        
    } // :: end GameScreen
    
    
}
        
        
        
